#!/usr/local/bin/perl5 -w
# -*- Perl -*-
#
# nmzcheck.pl - program to check index
#         by furukawa@tcp-ip.or.jp
#

open(STDERR, ">&STDOUT");
$| = 1;

use strict;
use IO::File;
use DirHandle;

my $dir = '.';

my $intsize = length pack('N', 0);

my $opt_c = 0;

while (@ARGV && $ARGV[0] =~ s/^-//){
    my $argv = shift(@ARGV);
    while ($argv =~ s/^(.)//){
        my $opt = $1;
        $opt_c = 1 if $opt eq 'c';
    }
}

push(@ARGV, '.') unless @ARGV;
&nmzcheck(shift(@ARGV)) while @ARGV;

sub nmzcheck{
    my $dir = shift;
    $dir =~ s/[\/]$//;
    my $dh = new DirHandle($dir);
    &error("Cannot open directory: $dir"), return unless defined $dh;
    &checkmain("$dir/NMZ.w", 'i', -1, \&get_line, \&check_w);
    &checkmain("$dir/NMZ.i", 'i', -1, \&get_block, \&check_packw);
    &checkmain("$dir/NMZ.p", 'i', 0x10000, \&get_block, \&check_packw);

    while (defined(my $ent = $dh->read)){
        if ($ent =~ /^NMZ\.field\.([^\.]+)$/){
            &checkmain("$dir/$ent", '.i', -1, \&get_line, \&check_field);
        }
    }
    $dh->close;
}

sub get_line{
    my $fh = shift;
    my $ptr = shift;
    $$ptr = $fh->getline;
    return defined($$ptr)? undef: "body read error";
}

sub get_block{
    my $fh = shift;
    my $ptr = shift;

    my $len = '';
    while (4 > length $len){
        my $c;
        last unless read($fh, $c, 1);
        $len .= $c;
        unless (0x80 & ord $c){
            $len = unpack('w', $len);
            return ($fh->read($$ptr, $len) && $len == length $$ptr)? undef: "body read error";
        }
    }
    return "body length error";
}

sub check_w{
    my $ptr = shift;
    return $$ptr =~ /^(([\xa1-\xfe]{2})+|[\x21-\x7e]+)\n$/? undef: "word error";
}

sub check_field{
    my $ptr = shift;
    return $$ptr =~ /^([\xa1-\xfe]{2}|[\x20-\x7e])*\n$/? undef: "string error";
}

sub check_packw{
    my $ptr = shift;
    eval {my @tmp = unpack('w*', $$ptr)};
    return $@;
}

sub checkmain{
    my $nb = shift;
    my $ni = $nb . shift;
    my $num = shift;
    my $get_body = shift;
    my $check_body = shift;

    print "Checking $nb\n";

    my $fb = new IO::File($nb);
    &error("Cannot open body"), return unless defined $fb;
    binmode $fb;

    my $fi = new IO::File($ni);
    &error("Cannot open index"), return unless defined $fi;
    binmode $fi;

    for (my $i = 0; $i != $num; $i++){
        my $n;
        my $error;
        my $body;
        &error("index read error"), return if !$fi->read($n, $intsize) || $intsize != length $n;
        next if ($n = unpack('N', $n)) == 0xffffffff && $num > 0;
        if ($fb->tell != $n){
            &error("offset error");
            return unless $opt_c;
        }
        if ($error = (&{$get_body}($fb, \$body) || &{$check_body}(\$body))){
            &error(sprintf("%08X(%d): %s", $n, $i, $error));
            return unless $opt_c;
        }
        last if eof($fi) && eof($fb);
    }
    &error("size error") unless eof($fi) && eof($fb);
}

sub error{
    print join("\n", @_) . "\n"
}

